import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyjak",
    version="0.1.1",
    author="Max Byrde",
    author_email="max.byrde@gmail.com",
    description="A Python library for reading and writing binary sources.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/miniwa/pyjak",
    license="MIT",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
)
